"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import threading
import sys
import numpy as np
from tqdm import tqdm
from numba import njit

from qrisp.circuit import QuantumCircuit, fast_append, XGate
from qrisp.simulator.circuit_preprocessing import (
    circuit_preprocessor,
    count_measurements_and_treat_alloc,
    group_qc,
    insert_multiverse_measurements,
)

from qrisp.simulator.quantum_state import QuantumState


# This functions determines the quantum state after executing a quantum circuit
# and afterwards extracts the probability of measuring certain bit strings
def run(qc, shots, token="", iqs=None, insert_reset=True):

    if len(qc.data) == 0:
        return {"": shots}
    if shots == 0:
        return {}

    progress_bar = tqdm(
        desc=f"Simulating {len(qc.qubits)} qubits..",
        bar_format="{desc} |{bar}| [{percentage:3.0f}%]",
        ncols=85,
        leave=False,
        delay=0.1,
        position=0,
        smoothing=1,
        file=sys.stdout,
    )

    LINE_CLEAR = "\x1b[2K"
    progress_bar.display()

    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append(2):

        qc = qc.transpile()

        # Count the amount of measurements (we can stop the simulation after all
        # measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(
            qc, insert_reset=insert_reset
        )

        # Apply circuit preprocessing more
        qc = circuit_preprocessor(qc)

        measurement_counter = 0

        for i in range(len(qc.data)):
            if qc.data[i].op.name == "measure":
                measurement_counter += 1
            if measurement_counter == measurement_amount:
                break

        qc.data = qc.data[: i + 1]

        # Counter to track how many measurements have been performed
        measurement_counter = 0

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object

        mes_list = []

        # if len(qc.qubits) < 30 or True:
        # qc, mes_list = extract_measurements(qc)

        qc, new_mes_list = insert_multiverse_measurements(qc)

        mes_list = mes_list + new_mes_list

        if iqs is None:
            # Create impure quantum state object. This object tracks multiple decoherent
            # quantum states that can appear when applying a non-unitary operation
            # iqs = ImpureQuantumState(len(qc.qubits), clbit_amount=len(qc.clbits))
            iqs = QuantumState(len(qc.qubits))

        mes_qubit_indices = []
        mes_clbit_indices = []

        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2 ** qc.data[i].op.num_qubits

        progress_bar.total = total_flops
        for i in range(len(qc.data)):
            # Set alias for the instruction of this operation
            instr = qc.data[i]
            progress_bar.update(2**instr.op.num_qubits)
            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions

            # Disentangling describes an operation, which mean that the superposition of
            # two states can be safely treated as two decoherent states. This is
            # advantageous because it might be possible that the amplitude of one
            # state is 0, which means that we halfed the workload. Even if both states
            # have non-zero amplitude, this still yields an improvement because
            # computing two decoherent states is more easily parallelized than the
            # combined coherent state
            if instr.op.name == "disentangle":
                # iqs.reset(qubit_indices[0], True)
                iqs.disentangle(qubit_indices[0], warning=instr.op.warning)

            # If the operation is unitary, we apply this unitary on to the required
            # qubit indices
            else:

                iqs.apply_operation(instr.op, qubit_indices)

            # If all measurements have been performed, break
            if measurement_counter == measurement_amount:
                break

        mes_list.sort(key=lambda x: -qc.clbits.index(x.clbits[0]))

        for instr in mes_list:
            mes_qubit_indices.append(qc.qubits.index(instr.qubits[0]))

        if len(mes_qubit_indices):
            outcome_list, cl_prob = iqs.multi_measure(
                mes_qubit_indices[::-1], return_res_states=False
            )
            mes_qubit_indices = []

        progress_bar.close()
        print("\r" + 85 * " ", end=LINE_CLEAR + "\r")

        # Prepare result dictionary
        # The iqs object contains the outcome bitstrings in the attribute .outcome_list
        # and the probablities in .cl_prob. In order to ensure qiskit compatibility, we
        # reverse the bitstrings
        cl_prob = np.round(cl_prob, int(-np.log10(np.max(cl_prob))) + 5)
        norm = np.sum(cl_prob)
        cl_prob = cl_prob / norm

        res = {}
        # If shots >= 1000000, no samples will be drawn and the distribution will
        # be returned instead
        if shots is None:

            for j in range(len(outcome_list)):

                outcome_str = bin(outcome_list[j])[2:].zfill(len(mes_list))

                p = float(cl_prob[j])

                if p == 0:
                    continue

                try:
                    res[outcome_str] += p
                except KeyError:
                    res[outcome_str] = p

        # Generate samples
        else:
            from numpy.random import choice

            # p_array = np.array(list(prob_dict.values()))

            # samples = choice(len(p_array), shots, p=p_array)

            samples = choice(len(cl_prob), shots, p=cl_prob)

            temp = gen_res_dict(samples)

            for k, v in temp.items():
                outcome_str = bin(outcome_list[k])[2:].zfill(len(mes_list))
                res[outcome_str] = int(v)

        return res


@njit
def gen_res_dict(samples):

    hist = np.histogram(samples, np.arange(np.max(samples) + 2))[0]

    temp = {}

    for i in range(len(hist)):

        if hist[i] != 0:
            temp[i] = hist[i]

    return temp


def statevector_sim(qc):

    progress_bar = tqdm(
        desc=f"Simulating {len(qc.qubits)} qubits..",
        bar_format="{desc} |{bar}| [{percentage:3.0f}%]",
        ncols=85,
        # ascii = " >#",
        leave=False,
        delay=0.2,
        position=0,
        smoothing=1,
        file=sys.stdout,
        # colour = "green"
    )

    LINE_CLEAR = "\x1b[2K"
    progress_bar.display()
    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append():
        # We convert the circuit which is given with portable objects to a qrisp circuit

        qc = qc.copy()

        # TO-DO fix qubit ordering
        # qc.qubits = qc.qubits[::-1]

        # Count the amount of measurements (we can stop the simulation after all
        # measurements are performed)
        measurement_amount = count_measurements_and_treat_alloc(qc, insert_reset=False)

        if measurement_amount != 0:
            raise Exception(
                "Tried to determine the statevector of a circuit containing a "
                "measurement"
            )

        # Apply circuit preprocessing more
        qc = group_qc(qc)

        if len(qc.data) == 0:
            res = np.zeros(2 ** len(qc.qubits), dtype=np.complex64)
            res[0] = 1

            progress_bar.close()
            print(LINE_CLEAR, end="\r")

            return res

        qs = QuantumState(len(qc.qubits))

        total_flops = 0
        for i in range(len(qc.data)):
            total_flops += 2 ** qc.data[i].op.num_qubits

        progress_bar.total = total_flops

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object
        for i in range(len(qc.data)):
            # Set alias for the instruction of this operation
            instr = qc.data[i]

            progress_bar.update(2**instr.op.num_qubits)
            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions
            qs.apply_operation(instr.op, qubit_indices)

        res = qs.eval().tensor_array.to_array()

        progress_bar.close()
        print("\r" + 85 * " ", end=LINE_CLEAR + "\r")

        # Deactivate the fast append mode
        QuantumCircuit.fast_append = False

        return res


def single_shot_sim(qc, quantum_state=None):
    if len(qc.data) == 0:
        return "", quantum_state

    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append():
        # We convert the circuit which is given with portable objects to a qrisp circuit

        # Treat allocation gates (ie. remove them)
        count_measurements_and_treat_alloc(qc, insert_reset=True)

        from qrisp.simulator import reorder_circuit

        qc = group_qc(qc)

        qc = reorder_circuit(qc, ["measure", "reset", "disentangle"])

        if quantum_state is None:
            # Create quantum state object.
            quantum_state = QuantumState(len(qc.qubits))

        result_str = len(qc.clbits) * ["0"]

        import random

        if len(qc.data):
            # Wrapper to pre-calculate the unitaries from the preprocessed circuit in
            # parallel with the main thread
            def pre_calc_unitaries(op):
                try:
                    op.get_unitary()
                except:
                    pass

            # pre_calc_unitaries()

            pre_calc_thr = threading.Thread(
                target=pre_calc_unitaries, args=(qc.data[0].op,)
            )
            pre_calc_thr.start()

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object
        for i in range(len(qc.data)):
            pre_calc_thr.join()

            if i < len(qc.data) - 1:
                pre_calc_thr = threading.Thread(
                    target=pre_calc_unitaries, args=(qc.data[i + 1].op,)
                )
                pre_calc_thr.start()

            # Set alias for the instruction of this operation
            instr = qc.data[i]

            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qc.qubits.index(qb) for qb in instr.qubits]

            # Perform instructions
            if instr.op.name == "reset":
                quantum_state.measure(qubit_indices[0])

                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome

                rng = random.random()

                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1

            # Disentangling describes an operation, which mean that the superposition of
            # two states can be safely treated as two decoherent states. This is
            # advantageous because it might be possible that the amplitude of one state
            # is 0, which means that we halfed the workload. Even if both states have
            # non-zero amplitude, this still yields an improvement because computing two
            # decoherent states is more easily parallelized than the combined coherent
            # state.

            elif instr.op.name == "measure":
                quantum_state.measure(qubit_indices[0])

                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome

                rng = random.random()

                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1
                    result_str[qc.clbits.index(instr.clbits[0])] = "1"

            elif instr.op.name[:4] == "c_if":
                if result_str[qc.clbits.index(instr.clbits[0])] == "1":
                    quantum_state.apply_operation(instr.op, qubit_indices)
            # If the operation is unitary, we apply this unitary on to the required qubit
            # indices
            else:
                quantum_state.apply_operation(instr.op, qubit_indices)

        return "".join(result_str)[::-1], quantum_state


def advance_quantum_state(qc, quantum_state, deallocated_qubits, qubit_to_index_dic):
    if len(qc.data) == 0:
        return quantum_state

    allocated_qubits = len(qc.qubits)
    max_req_qubits = allocated_qubits
    allocation_amount = 0

    for instr in qc.data:

        if instr.op.name == "qb_dealloc":
            allocated_qubits -= 1
            deallocated_qubits.append(instr.qubits[0])
        elif instr.op.name == "qb_alloc":
            allocated_qubits += 1
            allocation_amount += 1
            if allocated_qubits > max_req_qubits:
                max_req_qubits += 1

    progress_bar = tqdm(
        desc=f"Simulating {max_req_qubits-allocation_amount} qubits..",
        bar_format="{desc} |{bar}| [{percentage:3.0f}%]",
        ncols=85,
        leave=False,
        delay=0.2,
        position=0,
        smoothing=1,
        file=sys.stdout,
    )

    LINE_CLEAR = "\x1b[2K"
    progress_bar.display()

    # This command enables fast appending. Fast appending means that the .append method
    # of the QuantumCircuit class checks much less validity conditions and is also less
    # tolerant regarding inputs.
    with fast_append(2):
        # We convert the circuit which is given with portable objects to a qrisp circuit

        # Treat allocation gates (ie. remove them)
        count_measurements_and_treat_alloc(qc, insert_reset=True)
        qc = group_qc(qc)

        import random

        # Main loop - this loop successively executes operations onto the impure
        # quantum state object

        progress_bar.total = len(qc.data)

        # qubit_to_index_dic = {}
        # for i in range(len(qc.qubits)):
        #     qubit_to_index_dic[qc.qubits[i]] = i

        for i in range(len(qc.data)):

            # Set alias for the instruction of this operation
            instr = qc.data[i]

            progress_bar.update(1)

            # Gather the indices of the qubits from the circuits (i.e. integers instead
            # of the identifier strings)
            qubit_indices = [qubit_to_index_dic[qb] for qb in instr.qubits]

            # Perform instructions
            if instr.op.name == "reset":
                quantum_state.measure(qubit_indices[0])

                p_0, state_0, p_1, state_1 = quantum_state.last_mes_outcome

                rng = random.random()

                if rng < p_0:
                    quantum_state = state_0
                else:
                    quantum_state = state_1

            # Disentangling describes an operation, which mean that the superposition of
            # two states can be safely treated as two decoherent states. This is
            # advantageous because it might be possible that the amplitude of one state
            # is 0, which means that we halfed the workload. Even if both states have
            # non-zero amplitude, this still yields an improvement because computing two
            # decoherent states is more easily parallelized than the combined coherent
            # state.
            if instr.op.name == "disentangle":
                # iqs.reset(qubit_indices[0], True)
                quantum_state.disentangle(qubit_indices[0], warning=instr.op.warning)

            # If the operation is unitary, we apply this unitary on to the required qubit
            # indices
            else:
                quantum_state.apply_operation(instr.op, qubit_indices)

        progress_bar.close()
        print("\r" + 85 * " ", end=LINE_CLEAR + "\r")

        return quantum_state



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp import QuantumCircuit
from qrisp.alg_primitives.mcx_algs.circuit_library import toffoli_qc


# Function to synthesize a multi controlled X gate using gray synthesis
def gray_multi_cx(n):
    qs = QuantumCircuit(n + 1)
    if n == 1:
        qs.cx(qs.qubits[0], qs.qubits[1])
    elif n == 2:
        qs = toffoli_qc.copy()

    else:
        qs = QuantumCircuit(n + 1)
        qs.h(qs.qubits[-1])
        from qrisp.alg_primitives.logic_synthesis import GraySynthGate

        gsg = GraySynthGate((2 ** (n + 1) - 1) * [0] + [np.pi])
        qs.append(gsg, qs.qubits)
        qs.h(qs.qubits[-1])

    result = qs.to_gate()
    result.name = "gray multi cx"

    return result



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.core import ShapedArray
from qrisp.circuit import Reset, Qubit

from qrisp.jasp.primitives import (
    AbstractQuantumCircuit,
    AbstractQubit,
    QuantumPrimitive,
    AbstractQubitArray,
)

# Create the primitive
Measurement_p = QuantumPrimitive("measure")


@Measurement_p.def_abstract_eval
def measure_abstract_eval(meas_object, qc):
    """Abstract evaluation of the primitive.

    This function does not need to be JAX traceable. It will be invoked with
    abstractions of the actual arguments.
    Args:
      xs, ys, zs: abstractions of the arguments.
    Result:
      a ShapedArray for the result of the primitive.
    """

    if isinstance(meas_object, AbstractQubit):
        return ShapedArray((), bool), AbstractQuantumCircuit()
    elif isinstance(meas_object, AbstractQubitArray):
        return ShapedArray((), dtype="int64"), AbstractQuantumCircuit()
    else:
        raise Exception(
            f"Tried to call measurement primitive with type {type(meas_object)}"
        )


Measurement_p.multiple_results = True


@Measurement_p.def_impl
def measure_implementation(meas_object, qc):
    from qrisp import Qubit, QuantumCircuit

    return_bool = False
    if isinstance(meas_object, Qubit):
        meas_object = [meas_object]
        return_bool = True

    if isinstance(qc, QuantumCircuit):
        if return_bool:
            qc.measure(meas_object)
            return qc.clbits[-1], qc
        else:
            clbit_list = []
            for i in range(len(meas_object)):
                qc.measure(meas_object[i])
                clbit_list.append(qc.clbits[-1])
            return clbit_list, qc
    else:
        res = 0
        for i in range(len(meas_object)):
            res += 2**i * qc.measure([meas_object[i]])

        if return_bool:
            return bool(res), qc
        return res, qc


reset_p = QuantumPrimitive("reset")


@reset_p.def_abstract_eval
def reset_abstract_eval(reset_object, qc):
    """Abstract evaluation of the primitive.

    This function does not need to be JAX traceable. It will be invoked with
    abstractions of the actual arguments.
    Args:
      xs, ys, zs: abstractions of the arguments.
    Result:
      a ShapedArray for the result of the primitive.
    """
    return AbstractQuantumCircuit()


@reset_p.def_impl
def reset_implementation(reset_object, qc):
    if isinstance(reset_object, Qubit):
        reset_object = [reset_object]
    for i in range(len(reset_object)):
        qc.reset([reset_object[i]])
    return qc



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from qrisp.core import QuantumVariable, measure
from qrisp.jasp import sample

import jax
import jax.numpy as jnp
from jax.lax import while_loop, fori_loop


def get_jasp_measurement(
    hamiltonian,
    state_prep,
    state_args=(),
    precision=0.01,
    diagonalisation_method="commuting_qw",
):
    r"""
    This method returns the expected value of a Hamiltonian for the state of a quantum argument.

    Parameters
    ----------
    hamiltonian : QubitOperator
        The Hamiltonian for which the exptectaion value is measuered.
    state_prep : callable
        A function returning a QuantumVariable.
        The expectation of the Hamiltonian for the state from this QuantumVariable will be measured.
        The state preparation function can only take classical values as arguments.
        This is because a quantum value would need to be copied for each sampling iteration, which is prohibited by the no-cloning theorem.
    state_args : tuple
        A tuple of arguments of the ``state_prep`` function.
    precision: float, optional
        The precision with which the expectation of the Hamiltonian is to be evaluated.
        The default is 0.01. The number of shots scales quadratically with the inverse precision.
    diagonalisation_method : str, optional
        Specifies the method for grouping and diagonalizing the QubitOperator.
        Available are ``commuting_qw``, i.e., the operator is grouped based on qubit-wise commutativity of terms,
        and ``commuting``, i.e., the operator is grouped based on commutativity of terms.
        The default is ``commuting_qw``.

    Returns
    -------
    float
        The expected value of the Hamiltonian.

    """

    hamiltonian = hamiltonian.hermitize()
    hamiltonian = hamiltonian.eliminate_ladder_conjugates()
    hamiltonian = hamiltonian.apply_threshold(0)
    if len(hamiltonian.terms_dict) == 0:
        return 0

    if diagonalisation_method == "commuting_qw":
        temp_groups = hamiltonian.commuting_qw_groups()
        groups = []
        # In order for the change of basis function (below) to work properly,
        # the ladder terms either need to completely agree or completely disagree
        for group in temp_groups:
            groups.extend(
                group.group_up(
                    lambda a, b: a.ladders_agree(b) or not a.ladders_intersect(b)
                )
            )

    elif diagonalisation_method == "commuting":
        temp_groups = hamiltonian.group_up(lambda a, b: a.commute_pauli(b))
        groups = []
        # In order for the change of basis function (below) to work properly,
        # the ladder terms either need to completely agree or completely disagree
        for group in temp_groups:
            groups.extend(
                group.group_up(
                    lambda a, b: a.ladders_agree(b) or not a.ladders_intersect(b)
                )
            )

    else:
        raise Exception(f"Unknown diagonalisation method: {diagonalisation_method}.")

    samples = []
    meas_ops = []
    meas_coeffs = []
    stds = []

    # Compute amounts of shots
    for group in groups:
        # Collect standard deviation
        n = hamiltonian.find_minimal_qubit_amount()
        stds.append(np.sqrt(group.get_operator_variance(n=n)))

    N = sum(stds)
    shots_list = [N * s for s in stds]

    for index, group in enumerate(groups):

        # Calculate the new measurement operators (after change of basis)
        meas_op = group.change_of_basis(method=diagonalisation_method)

        def new_state_prep(state_args):
            qv = state_prep(*state_args)
            group.change_of_basis(qv, method=diagonalisation_method)
            return qv

        shots = int(shots_list[index] / precision**2)
        res = sample(new_state_prep, shots=shots)(state_args)

        samples.append(jnp.int64(res))

        temp_meas_ops = []
        temp_coeff = []
        for term, coeff in meas_op.terms_dict.items():
            temp_meas_ops.append(term.serialize())
            temp_coeff.append(coeff)

        meas_coeffs.append(temp_coeff)
        meas_ops.append(temp_meas_ops)

    expectation = jasp_evaluate_expectation_jitted(samples, meas_ops, meas_coeffs)
    return expectation


@jax.jit
def jasp_evaluate_expectation_jitted(samples, operators, coefficients):
    """
    Evaluate the expectation.

    """

    def body_fun(i, val):
        expectation, N, op, samples, coefficient = val
        expectation += (
            1
            / N
            * jasp_evaluate_observable_jitted(op, samples[i])
            * jnp.real(coefficient)
        )
        return expectation, N, op, samples, coefficient

    expectation = 0

    for index1, ops in enumerate(operators):
        for index2, op in enumerate(ops):
            N = len(samples[index1])
            expectation, _, _, _, _ = fori_loop(
                0,
                N,
                body_fun,
                (expectation, N, op, samples[index1], coefficients[index1][index2]),
            )

    return expectation


@jax.jit
def jasp_evaluate_observable_jitted(observable: tuple, x: int):
    # This function evaluates how to compute the energy of a measurement sample x.
    # Since we are also considering ladder operators, this energy can either be
    # 0, -1 or 1. For more details check out the comments of QubitOperator.get_conjugation_circuit

    # The observable is given as tuple, containing four integers.
    # To understand the meaning of these integers check QubitTerm.serialize.

    # Unwrap the tuple
    z_int, AND_bits, AND_ctrl_state, contains_ladder = observable

    # Compute whether the sign should be sign flipped based on the Z operators
    sign_flip_int = z_int & x
    sign_flip = 0

    def cond_fun(state):
        sign_flip_int, sign_flip = state
        return sign_flip_int > 0

    def body_fun(state):
        sign_flip_int, sign_flip = state
        sign_flip += sign_flip_int & 1
        sign_flip_int >>= 1
        return sign_flip_int, sign_flip

    sign_flip_int, sign_flip = while_loop(
        cond_fun, body_fun, (sign_flip_int, sign_flip)
    )

    # If there is a ladder operator in the term, we need to half the energy
    # because we want to measure (|110><110| - |111><111|)/2
    prefactor = 1 - 0.5 * contains_ladder

    # If there are no AND bits, we return the result
    # Otherwise we apply the AND_ctrl_state to flip the appropriate bits.
    corrected_x = x ^ AND_ctrl_state

    # If all bits are in the 0 state the AND is true.
    return (
        prefactor
        * (-1) ** sign_flip
        * jnp.int64((AND_bits == 0) | (corrected_x & AND_bits == 0))
    )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumFloat, conjugate, measure
from qrisp.jasp import make_jaspr, RUS
from qrisp.alg_primitives.switch_case import qswitch
from qrisp.algorithms.grover.grover_tools import tag_state
from qrisp.alg_primitives.amplitude_amplification import amplitude_amplification
import numpy as np


def inner_LCU(operand_prep, state_prep, unitaries, num_unitaries=None, oaa_iter=0):
    r"""
    Core implementation of the Linear Combination of Unitaries (LCU) protocol without
    Repeat-Until-Success (RUS) protocol. The LCU method is a foundational quantum algorithmic
    primitive that enables the application of a non-unitary operator $A$, expressed as a weighted
    sum of unitaries $U_i$ as $A=\sum_i\alpha_i U_i$, to a quantum state, by embedding $A$ into a larger unitary circuit. This is
    central to quantum algorithms for `Hamiltonian simulation <https://www.taylorfrancis.com/chapters/edit/10.1201/9780429500459-11/simulating-physics-computers-richard-feynman>`_, `Linear Combination of Hamiltonian Simulation (LCHS) <https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.131.150603>`_, Quantum Linear Systems (e.g. `HHL algorithm <https://pennylane.ai/qml/demos/linear_equations_hhl_qrisp_catalyst>`_), `Quantum Signal
    Processing (QSP) <https://journals.aps.org/prxquantum/abstract/10.1103/PRXQuantum.5.020368>`_, and `Quantum Singular Value Transformation (QSVT) <https://dl.acm.org/doi/abs/10.1145/3313276.3316366>`_.

    This function implements the prepare-select-unprepare structure, also known as block encoding:

    .. math::
        \mathrm{LCU} = \mathrm{PREPARE}^\dagger \cdot \mathrm{SELECT} \cdot \mathrm{PREPARE}

    - **PREPARE**: Prepares an ancilla quantum variable in a superposition encoding the normalized coefficients $\alpha_i\geq0$ of the target operator

    .. math ::

            \mathrm{PREPARE}|0\rangle=\sum_i\sqrt{\frac{\alpha_i}{\lambda}}|i\rangle

    - **SELECT**: Applies the unitary $U_i$ to the input state $\ket{\psi}$, controlled on the ancilla variable being in state $|i\rangle$.

    .. math ::

        \mathrm{SELECT}|i\rangle|\psi\rangle=|i\rangle U_i|\psi\rangle

    - **PREPARE**$^\dagger$: Applies the inverse prepartion to the ancilla.

    .. note::

        The LCU protocol is deemed successful only if the ancilla variable is measured in the $\ket{0}$ state, which occurs with a probability proportional to :math:`\frac{\langle\psi|A^{\dagger}A|\psi\rangle}{\lambda^2}` where $\lambda=\sum_i\alpha_i$.
        This function does not perform the measurement; it returns the ancilla variable and the transformed target variable.

    For a complete implementation of LCU with the Repeat-Until-Success protocol, see :func:`LCU`.

    For more details on the LCU protocol, refer to `Childs and Wiebe (2012) <https://arxiv.org/abs/1202.5822>`_, or `related seminars provided by Nathan Wiebe <https://www.youtube.com/watch?v=irMKrOIrHP4>`_.

    Parameters
    ----------
    operand_prep : callable
        A function preparing the input state $\ket{\psi}$. This function must return a :ref:`QuantumVariable` (the ``operand``).
    state_prep : callable
        A function preparing the coefficient state from the $\ket{0}$ state.
        This function receives a :ref:`QuantumFloat` with $\lceil\log_2m\rceil$ qubits for $m$ unitiaries $U_0,\dotsc,U_{m-1}$ as argument and applies

        .. math ::

            \text{PREPARE}\ket{0} = \sum_i\sqrt{\frac{\alpha_i}{\lambda}}\ket{i}

    unitaries : list[callable] or callable
        Either:
          - A list of functions performing some in-place operation on ``operand``, or
          - A function ``unitaries(i, operand)`` performing some in-place operation on ``operand`` depending on a nonnegative integer index ``i`` specifying the case.

    num_unitaries : int, optional
        Required when ``unitaries`` is a callable to specify the number $m$ of unitaries.
    oaa_iter : int, optional
        The number of iterations of oblivious amplitude amplification to perform. The default is 0.

    Returns
    -------
    tuple(:ref:`QuantumFloat`, :ref:`QuantumVariable`)
          - **case_indicator** : Ancilla variable encoding which unitary was selected.
          - **operand** : Target quantum variable after the LCU operation.

    Raises
    ------
    TypeError
        If ``unitaries`` is not a list or a callable.
    ValueError
        If ``num_unitaries`` is not specified when ``unitaries`` is a callable.

    See Also
    --------
    LCU : Full LCU implementation using the RUS protocol.
    view_LCU : Generates the quantum circuit for visualization.

    Examples
    --------

    As a **first example**, we apply the non-unitary operator $A$ to the operand $\ket{\psi}=\ket{1}$ where

    .. math::

        A = \begin{pmatrix}1 & 1\\ 1 & 1\end{pmatrix} = \begin{pmatrix}1 & 0\\ 0 & 1\end{pmatrix} + \begin{pmatrix}0 & 1\\ 1 & 0\end{pmatrix} = I + X

    That is,

    .. math:: A = \alpha_0U_0 + \alpha_1U_1

    where $\alpha_0=\alpha_1=1$, and $U_0=I$, $U_1=X$.

    Accordingly, we define the unitaries

    ::

        from qrisp import *

        def U0(operand):
            pass

        def U1(operand):
            x(operand)

        unitaries = [U0, U1]

    and the ``state_prep`` function implementing

    .. math ::

        \text{PREPARE}\ket{0} = \frac{1}{\sqrt{2}}\left(\ket{0}+\ket{1}\right)

    ::

        def state_prep(case):
            h(case)

    Next, we define the ``operand_prep`` function preparing the state $\ket{\psi}=\ket{1}$

    ::

        def operand_prep():
            operand = QuantumVariable(1)
            x(operand)
            return operand

    Finally, we apply ``inner_LCU``

    >>> case_indicator, operand = inner_LCU(operand_prep, state_prep, unitaries)
    >>> operand.qs.statevector()

    As result we obtain the state

    .. math ::

        \frac12((\ket{0}+\ket{1})\ket{0}_{\text{case}} - \frac12(\ket{0}-\ket{1})\ket{1}_{\text{case}}

    If we now measure the ``case_indicator`` in the state $\ket{0}$, the operand will be in state $\frac{1}{\sqrt{2}}(\ket{0}+\ket{1})$
    implementing (up to rescaling) the non-unitary operator $A$ acting on the input state $\ket{\psi}=\ket{1}$.

    """

    operand = operand_prep()

    if not callable(unitaries):
        if not isinstance(unitaries, list):
            raise TypeError("unitaries must be callable or list[callable].")
        num_unitaries = len(unitaries)
    else:
        if num_unitaries == None:
            raise ValueError(
                "The number of unitiaries must be specified if unitaries is callable."
            )

    # Specify the QunatumVariable that indicates which case to execute
    n = np.int64(np.ceil(np.log2(num_unitaries)))
    case_indicator = QuantumFloat(n)

    # LCU protocol with conjugate preparation
    def LCU_state_prep(case_indicator, operand):
        with conjugate(state_prep)(case_indicator):
            qswitch(operand, case_indicator, unitaries)

    def oracle_func(case_indicator, operand):
        tag_state({case_indicator: 0})

    LCU_state_prep(case_indicator, operand)

    if oaa_iter > 0:
        amplitude_amplification(
            [case_indicator, operand],
            LCU_state_prep,
            oracle_func,
            reflection_indices=[0],
            iter=oaa_iter,
        )

    return case_indicator, operand


def LCU(operand_prep, state_prep, unitaries, num_unitaries=None, oaa_iter=0):
    r"""
    Full implementation of the Linear Combination of Unitaries (LCU) algorithmic primitive using the
    Repeat-Until-Success (RUS) protocol.

    This function constructs and executes the LCU protocol using the provided state preparation function
    and unitaries. It utilizes the :func:`qrisp.jasp.RUS` decorator from Jasp to handle the repeat-until-success
    mechanism, which repeatedly applies the LCU operation until the ancilla register is measured in the $|0\rangle$ state, indicating a successful implementation. The LCU algorithm enables the implementation of linear combinations of unitary operations
    on a quantum variable by probabilistically projecting onto the desired transformation. The terminal_sampling decorator is utilized to evaluate the LCU.

    For more details on the LCU primitive, refer to `Childs and Wiebe (2012) <https://arxiv.org/abs/1202.5822>`_.

    For more information on the inner workings of this LCU implementation, see :func:`inner_LCU`.

    Parameters
    ----------
    operand_prep : callable
        A function preparing the input state $\ket{\psi}$. This function must return a :ref:`QuantumVariable` (the ``operand``).
    state_prep : callable
        A function preparing the coefficient state from the $\ket{0}$ state.
        This function receives a :ref:`QuantumFloat` with $\lceil\log_2m\rceil$ qubits for $m$ unitiaries $U_0,\dotsc,U_{m-1}$ as argument and applies

        .. math ::

            \text{PREPARE}\ket{0} = \sum_i\sqrt{\frac{\alpha_i}{\lambda}}\ket{i}

    unitaries : list[callable] or callable
        Either:
          - A list of functions performing some in-place operation on ``operand``, or
          - A function ``unitaries(i, operand)`` performing some in-place operation on ``operand`` depending on a nonnegative integer index ``i`` specifying the case.

    num_unitaries : int, optional
        Required when ``unitaries`` is a callable to specify the number $m$ of unitaries.
    oaa_iter : int, optional
        The number of iterations of oblivious amplitude amplification to perform. The default is 0.

    Returns
    -------
    :ref:`QuantumVariable`
        A variable representing the output state $A\ket{\psi}$ after successful application of LCU to the input state $\ket{\psi}$.

    Raises
    ------
    TypeError
        If ``unitaries`` is not a list or callable.
    ValueError
        If ``num_unitaries`` is not specified when ``unitaries`` is a callable.

    See Also
    --------
    inner_LCU : Core LCU implementation without RUS.
    view_LCU : Generates the quantum circuit for visualization.

    Examples
    --------

    As a **first example**, we apply the non-unitary operator $A$ to the operand $\ket{\psi}=\ket{1}$ where

    .. math::

        A = \begin{pmatrix}1 & 1\\ 1 & 1\end{pmatrix} = \begin{pmatrix}1 & 0\\ 0 & 1\end{pmatrix} + \begin{pmatrix}0 & 1\\ 1 & 0\end{pmatrix} = I + X

    This is,

    .. math:: A = \alpha_0U_0 + \alpha_1U_1

    where $\alpha_0=\alpha_1=1$, and $U_0=I$, $U_1=X$.

    Accordingly, we define the unitaries

    ::

        from qrisp import *

        def U0(operand):
            pass

        def U1(operand):
            x(operand)

        unitaries = [U0, U1]

    and the ``state_prep`` function implementing

    .. math ::

        \text{PREPARE}\ket{0} = \frac{1}{\sqrt{2}}\left(\ket{0}+\ket{1}\right)

    ::

        def state_prep(case):
            h(case)

    Next, we define the ``operand_prep`` function preparing the state $\ket{\psi}=\ket{1}$

    ::

        def operand_prep():
            operand = QuantumVariable(1)
            x(operand)
            return operand

    Finally, we apply LCU

    ::

        @terminal_sampling
        def main():

            qv = LCU(operand_prep, state_prep, unitaries)
            return qv

    and simulate

    >>> main()
    {0: 0.5, 1: 0.5}


    As a **second example**, we apply the operator

    .. math::

        \cos(H) = \frac{e^{iH}+e^{-iH}}{2}

    for some :ref:`Hermitian operator <operators>` $H$ to the input state $\ket{\psi}=\ket{0}$.

    First, we define an operator $H$ and unitaries performing the Hamiltonian evolutions $e^{iH}$ and $e^{-iH}$.
    (In this case, Trotterization will perform Hamiltonian evolution exactly since the individual terms commute.)

    ::

        from qrisp import *
        from qrisp.operators import X,Y,Z

        H = Z(0)*Z(1) + X(0)*X(1)

        def U0(operand):
            H.trotterization(forward_evolution=False)(operand)

        def U1(operand):
            H.trotterization(forward_evolution=True)(operand)

        unitaries = [U0, U1]

    Next, we define the ``state_prep`` and ``operand_prep`` functions

    ::

        def state_prep(case):
            h(case)

        def operand_prep():
            operand = QuantumVariable(2)
            return operand

    Finally, we apply LCU

    ::

        @terminal_sampling
        def main():

            qv = LCU(operand_prep, state_prep, unitaries)
            return qv

    and simulate

    >>> main()
    {3: 0.85471756539818, 0: 0.14528243460182003}

    Let's compare to the classically calculated result:

    >>> A = H.to_array()
    >>> from scipy.linalg import cosm
    >>> print(cosm(A))
    [[ 0.29192658+0.j  0.        +0.j  0.        +0.j -0.70807342+0.j]
    [ 0.        +0.j  0.29192658+0.j  0.70807342+0.j  0.        +0.j]
    [ 0.        +0.j  0.70807342+0.j  0.29192658+0.j  0.        +0.j]
    [-0.70807342+0.j  0.        +0.j  0.        +0.j  0.29192658+0.j]]

    That is, starting in state $\ket{\psi}=\ket{0}=(1,0,0,0)$, we obtain

    >>> result = cosm(A)@(np.array([1,0,0,0]).transpose())
    >>> result = result/np.linalg.norm(result) # normalise
    >>> result = result**2 # compute measurement probabilities
    >>> print(result)
    [0.1452825+0.j 0.       +0.j 0.       +0.j 0.8547175-0.j]

    which are exactly the probabilities we obsered in the quantum simulation!

    """

    case_indicator, qv = inner_LCU(
        operand_prep, state_prep, unitaries, num_unitaries, oaa_iter
    )

    # Success condition
    success_bool = measure(case_indicator) == 0
    return success_bool, qv


# Apply the RUS decorator with the workaround in order to show in documentation
temp_docstring = LCU.__doc__
LCU = RUS(static_argnums=[3, 4])(LCU)
LCU.__doc__ = temp_docstring


def view_LCU(operand_prep, state_prep, unitaries, num_unitaries=None):
    r"""
    Generate and return the quantum circuit for the LCU algorithm without utilizing
    the Repeat-Until-Success (RUS) protocol.

    This function constructs the LCU primitive and returns the corresponding quantum
    circuit representation. It's useful for visualization and analysis of the
    underlying quantum circuit implementing the LCU algorithm.

    Parameters
    ----------
    operand_prep : callable
        A function preparing the input state $\ket{\psi}$. This function must return a :ref:`QuantumVariable` (the ``operand``).
    state_prep : callable
        A function preparing the coefficient state from the $\ket{0}$ state.
        This function receives a :ref:`QuantumFloat` with $\lceil\log_2m\rceil$ qubits for $m$ unitiaries $U_0,\dotsc,U_{m-1}$ as argument and applies

        .. math ::

            \text{PREPARE}\ket{0} = \sum_i\sqrt{\frac{\alpha_i}{\lambda}}\ket{i}

    unitaries : list[callable] or callable
        Either:
          - A list of functions performing some in-place operation on ``operand``, or
          - A function ``unitaries(i, operand)`` performing some in-place operation on ``operand`` depending on a nonnegative integer index ``i`` specifying the case.

    num_unitaries : int, optional
        Required when ``unitaries`` is a callable to specify the number $m$ of unitaries.

    Returns
    -------
    :ref:`QuantumCircuit`
        Quantum circuit object showing the LCU implementation details.

    See Also
    --------
    inner_LCU : Core LCU implementation.
    LCU : Full LCU implementation using the RUS protocol.

    Examples
    --------

    ::

        from qrisp import *

        def U0(operand):
        y(operand)

        def U1(operand):
            x(operand)

        unitaries = [U0, U1]

        def state_prep(case):
            h(case)

        def operand_prep():
            operand = QuantumVariable(1)
            y(operand)
            return operand

        qc = view_LCU(operand_prep, state_prep, unitaries)

    >>> print(qc)
            ┌───┐                     ┌───┐                        »
    qb_287: ┤ Y ├─────────────────────┤ Y ├────────────────────────»
            ├───┤┌───────────────────┐└─┬─┘┌──────────────────────┐»
    qb_288: ┤ H ├┤0                  ├──┼──┤0                     ├»
            └───┘│  jasp_balauca_mcx │  │  │  jasp_balauca_mcx_dg │»
    qb_289: ─────┤1                  ├──■──┤1                     ├»
                 └───────────────────┘     └──────────────────────┘»
    «                             ┌───┐
    «qb_287: ─────────────────────┤ X ├─────────────────────────────
    «        ┌───────────────────┐└─┬─┘┌──────────────────────┐┌───┐
    «qb_288: ┤0                  ├──┼──┤0                     ├┤ H ├
    «        │  jasp_balauca_mcx │  │  │  jasp_balauca_mcx_dg │└───┘
    «qb_289: ┤1                  ├──■──┤1                     ├─────
    «        └───────────────────┘     └──────────────────────┘

    We can see that the operand and case variables are prepared, the unitaries for the two cases (i.e., X and Y) are executed, and the inverse preparation (H gate) of the case variable is applied.
    The inner workings of the circuit can be further analyzed by calling ``qc.transpile(level: int)``.

    """

    if not callable(unitaries):
        if not isinstance(unitaries, list):
            raise TypeError("unitaries must be callable or list[callable].")
        num_unitaries = len(unitaries)
    else:
        if num_unitaries == None:
            raise ValueError(
                "The number of unitiaries must be specified if unitaries is callable."
            )

    jaspr = make_jaspr(inner_LCU)(operand_prep, state_prep, unitaries, num_unitaries)

    # Convert Jaspr to quantum circuit and return the circuit
    return jaspr.to_qc(num_unitaries)[-1].transpile(3)
